// SPDX-License-Identifier: MIT

use askama::Template;
use axum::{
    extract::Form,
    response::{IntoResponse, Redirect},
    routing::{get, post},
    Router,
};
use serde::Deserialize;

use crate::common::{
    self, html_template::HtmlTemplate, CheckAuth, IptablesInfo, NatInfo, PortForwardingInfo,
};
use crate::error::PageResult;

#[derive(Deserialize, Debug)]
struct NatDelParam {
    delete_interface: String,
}

#[derive(Deserialize, Debug)]
struct PortForwardingDelParam {
    delete_pf: String,
}

#[derive(Template)]
#[template(path = "../src/nat/templates/nat.html")]
struct NatTemplate {
    nat_list: Vec<NatInfo>,
    pf_list: Vec<PortForwardingInfo>,
    network_interfaces: Vec<String>,
}

pub trait NatRoutes {
    fn nat_routes(self) -> Self;
}

impl NatRoutes for Router {
    fn nat_routes(self) -> Self {
        self.route("/nat", get(nat))
            .route("/nat_setup", post(nat_setup))
            .route("/nat_delete", post(nat_delete))
            .route("/port_forwarding_setup", post(port_forwarding_setup))
            .route("/port_forwarding_delete", post(port_forwarding_delete))
    }
}

async fn nat(_loggedin: CheckAuth) -> PageResult {
    let iptables_info = IptablesInfo::get().await?;
    let network_interfaces = common::network_interfaces();

    let template = NatTemplate {
        nat_list: iptables_info.nat,
        pf_list: iptables_info.port_forwarding,
        network_interfaces,
    };
    Ok(HtmlTemplate::new("NAT設定", template).into_response())
}

async fn nat_setup(_loggedin: CheckAuth, Form(nat_param): Form<NatInfo>) -> PageResult {
    let args = &["nat_setup.sh", &nat_param.out_interface];
    common::exec_command(args).await?;

    Ok(Redirect::to("/nat").into_response())
}

async fn nat_delete(_loggedin: CheckAuth, Form(nat_del_param): Form<NatDelParam>) -> PageResult {
    let args = &["nat_delete.sh", &nat_del_param.delete_interface];
    common::exec_command(args).await?;

    Ok(Redirect::to("/nat").into_response())
}

async fn port_forwarding_setup(
    _loggedin: CheckAuth,
    Form(pf_param): Form<PortForwardingInfo>,
) -> PageResult {
    let args = &[
        "port_forwarding_setup.sh",
        &pf_param.interface,
        &pf_param.protocol,
        &pf_param.dport,
        &pf_param.destination,
        &pf_param.destination_port,
    ];
    common::exec_command(args).await?;

    Ok(Redirect::to("/nat").into_response())
}

async fn port_forwarding_delete(
    _loggedin: CheckAuth,
    Form(pf_del_param): Form<PortForwardingDelParam>,
) -> PageResult {
    let mut args = vec!["port_forwarding_delete.sh"];
    for param in pf_del_param.delete_pf.split(',') {
        args.push(param)
    }
    common::exec_command(&args).await?;

    Ok(Redirect::to("/nat").into_response())
}
